/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Focus
Version:   1.0
ID:        ID_FOCUS
Category:  GUI
Date:      June 2003
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2000-2003.  All rights reserved.
Keywords:  focus, lostfocus, activate, userfocus
Short:     The Focus class manages user focussing.

###DESCRIPTION###
<p>This class is designed to aid the handling of the 'user focus' on
graphical areas.  As the user moves the mouse pointer around the display and
clicks on the various rendered objects, the system automatically records the
object that has the current focus, and notifies those that do not.  If you're
in a situation where you need to respond to the focus being lost or received,
then you'll need to use the focus class.</p>

<p>In order to create a functional focus object, all you need to do is set
the Object field to point to the object that you are interested in monitoring.
By default, the Focus object will only take action when the focus is received.
If you want to take action when the focus is lost, set the Lost field to TRUE.
When the focus object receives a positive notification of change, it will
do two things:</p>

<ul>
<li>If the focus object is contained by a Render object and the Frame field
has been set, then the frame of that Render object will be updated.</li>
<li>All children of the focus object will be sent Activate messages to
perform whatever commands you specified.</li>
</ul>

<p>The following example creates a render object that has the focus, then
self-destructs when the focus is lost (which will occur when the user clicks
on something else):</p>

<pre>
  &lt;render name="drawable" x="10" y="10" width="100" height="100"&gt;
    &lt;focus object="[owner]" lost&gt;
      &lt;action static call="free" object="[drawable]"/&gt;
    &lt;/focus&gt;
    &lt;action call="focus" object="[owner]"/&gt;
  &lt;render&gt;
</pre>

<p>If you require more examples, the best idea is to search some of the
existing DML files that are distributed with Athene.</p>

###END###

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/system/all.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_FOCUS  1.0

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_FOCUS, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2003.  All rights reserved.",
   "June 2003",
   "Focus"
};

static struct KernelBase  *KernelBase;
static OBJECTPTR FocusClass = 0;

/*****************************************************************************
** Focus class definition.
*/

struct Focus {
   OBJECT_HEADER          /* Class header */
   OBJECTID ObjectID;     /* Object to monitor */
   LONG Frame;            /* The frame to change the Render object to when focus event occurs */
   LONG LostFrame;        /* Frame to use after the focus event is reversed */
   LONG Lost;             /* Set to TRUE if activity should occur when the focus is lost */

   /*** Private fields ***/
   OBJECTID DrawableID;
   LONG PreviousFrame;
};

static struct FieldArray FocusFields[] = {
   { "Object",    0, FDF_OBJECTID|FDF_RW, 0, NULL, NULL },
   { "Frame",     0, FDF_LONG|FDF_RW,     0, NULL, NULL },
   { "LostFrame", 0, FDF_LONG|FDF_RW,     0, NULL, NULL },
   { "Lost",      0, FDF_LONG|FDF_RW,     0, NULL, NULL },
   END_FIELD
};

static ERROR FOCUS_ActionNotify(struct Focus *, struct acActionNotify *);
static ERROR FOCUS_ClosingTag(struct Focus *, APTR);
static ERROR FOCUS_Free(struct Focus *, APTR);
static ERROR FOCUS_NewObject(struct Focus *, APTR);
static ERROR FOCUS_Init(struct Focus *, APTR);

static struct ActionArray FocusActions[] = {
   { AC_ActionNotify, FOCUS_ActionNotify },
   { AC_ClosingTag,   FOCUS_ClosingTag },
   { AC_Free,         FOCUS_Free },
   { AC_NewObject,    FOCUS_NewObject },
   { AC_Init,         FOCUS_Init },
   { NULL, NULL }
};

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&FocusClass, NULL,
      FID_BaseClassID|TLONG, ID_FOCUS,
      FID_Name|TSTRING,      "Focus",
      FID_Category|TLONG,    CCF_GUI,
      FID_Actions|TPTR,      FocusActions,
      FID_Fields|TPTR,       FocusFields,
      FID_Size|TLONG,        sizeof(struct Focus),
      FID_Version|TFLOAT,    VER_FOCUS,
      TAGEND));
}


/*****************************************************************************
** Command: Expunge()
*/

ERROR CMDExpunge(void)
{
   if (FocusClass) { acFree(FocusClass); FocusClass = 0; }
   return(ERR_Okay);
}

/*****************************************************************************
** Focus: ActionNotify
*/

static ERROR FOCUS_ActionNotify(struct Focus *Self, struct acActionNotify *NotifyArgs)
{
   #define MAX_OBJECTS 20
   OBJECTPTR drawable;
   struct ChildEntry list[MAX_OBJECTS];
   LONG i, count;

   if (NotifyArgs->ActionID IS AC_Focus) {
      /*** Update the rendered frame if required ***/

      if ((Self->Frame) AND (Self->DrawableID)) {
         if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
            GetField(drawable, FID_Frame, FT_LONG, &Self->PreviousFrame);
            SetField(drawable, FID_Frame, FT_LONG, &Self->Frame);
            DelayMsg(AC_Draw, Self->DrawableID, NULL);
            ReleaseObject(drawable);
         }
      }

      /*** Activate any children that belong to us ***/

      if (Self->Lost IS FALSE) {
         count = MAX_OBJECTS;
         if (ListChildren(GetUniqueID(Self), list, &count) IS ERR_Okay) {
            for (i=0; i < count; i++) {
               acActivateID(list[i].ObjectID);
            }
         }
      }
   }
   else if (NotifyArgs->ActionID IS AC_LostFocus) {
      struct acUserClickRelease *Release;

      Release = (struct acUserClickRelease *)NotifyArgs->Args;

      /*** Change the render object's graphical frame if necessary ***/

      if ((Self->DrawableID) AND ((Self->LostFrame != -1) OR (Self->PreviousFrame != -1))) {
         if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
            if (Self->LostFrame != -1) SetField(drawable, FID_Frame, FT_LONG, &Self->LostFrame);
            else SetField(drawable, FID_Frame, FT_LONG, &Self->PreviousFrame);
            DelayMsg(AC_Draw, Self->DrawableID, NULL);
            ReleaseObject(drawable);
         }
      }

      /*** Activate any children that belong to us ***/

      if (Self->Lost IS TRUE) {
         count = MAX_OBJECTS;
         if (ListChildren(GetUniqueID(Self), list, &count) IS ERR_Okay) {
            for (i=0; i < count; i++) {
               acActivateID(list[i].ObjectID);
            }
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Focus: ClosingTag
*/

static ERROR FOCUS_ClosingTag(struct Focus *Self, APTR Void)
{
   return(ERR_Okay);
}

/*****************************************************************************
** Focus: Free
*/

static ERROR FOCUS_Free(struct Focus *Self, APTR Void)
{
   OBJECTPTR object;

   /*** Remove all subscriptions ***/

   if (Self->ObjectID) {
      if (AccessObject(Self->ObjectID, 3000, &object) IS ERR_Okay) {
         UnsubscribeAction(object, NULL, GetUniqueID(Self));
         ReleaseObject(object);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Focus: Init
*/

static ERROR FOCUS_Init(struct Focus *Self, APTR Void)
{
   OBJECTPTR Object;
   #define MAX_OBJECTS 20
   struct ChildEntry list[MAX_OBJECTS];
   OBJECTID OwnerID;
   LONG i, count;

   /* Check if we are associated with a Render object, whether
   ** it be a child or owner.
   */

   if (!Self->DrawableID) {
      count = MAX_OBJECTS;
      if (ListChildren(GetUniqueID(Self), list, &count) IS ERR_Okay) {
         for (i=0; i < count; i++) {
            if (list[i].ClassID IS ID_RENDER) {
               Self->DrawableID = list[i].ObjectID;
               break;
            }
         }
      }

      if (!Self->DrawableID) {
         OwnerID = GetOwner(Self);
         while ((OwnerID) AND (GetClassID(OwnerID) != ID_RENDER)) {
            OwnerID = GetOwnerID(OwnerID);
         }
         if (OwnerID) Self->DrawableID = OwnerID;
         else return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
      }
   }

   /* Subscribe to the focus and lostfocus actions of the object that
   ** we are supposed to be monitoring.
   */

   if (!Self->ObjectID) Self->ObjectID = Self->DrawableID;

   if (AccessObject(Self->ObjectID, 4000, &Object) IS ERR_Okay) {
      SubscribeActionTags(Object, Self, AC_Focus, AC_LostFocus, TAGEND);
      ReleaseObject(Object);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Focus: NewObject
*/

static ERROR FOCUS_NewObject(struct Focus *Self, APTR Void)
{
   Self->Lost = FALSE;
   Self->LostFrame = -1;
   Self->PreviousFrame = -1;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Drawable
Short:  Determines the drawable to use for frame setting.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>In most cases, when you create a Focus object it should be contained by a
drawable area, such as a Render object.  If this is not possible then you
will need to set this field to point to the drawable object that you want
to monitor.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Frame
Short:  Setting this field allows the object to alter rendered frames.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>You may want to use the focus object to provide a graphical response to
the focus being received or lost.  If so, set this field to the number of the
frame that you wish to draw to when the focus is changed.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Lost
Short:  Set to TRUE if the focus object should pay attention to focus loss.
Type:   BOOLEAN
Status: Read/Write

###DESCRIPTION###
<p>If you set this field to TRUE then the focus object will reverse its
behaviour and perform only when the focus is lost, rather than when the
focus is received.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Object
Short:  Set this field to determine what object is to be monitored for the focus.
Type:   OBJECTID
Status: Read/Write

###DESCRIPTION###
<p>Before initialising a Focus object, you should first set the Object field
so that the correct object is monitored for focus events.  If you do not set
this field, the Focus object will monitor the container that is active at
the time of initialisation.</p>

###END###

*****************************************************************************/
