/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      FileView
Version:   1.0
ID:        ID_FILEVIEW
Status:    Stable
Category:  FileSystem
Date:      March 2004
Author:    Rocklyte Systems
Copyright: Rocklyte Systems, 2000-2004.  All rights reserved.
Keywords:  File, Directory, Manager, Files, Directories, Management, Explorer
Short:     The FileView class provides a GUI-based gateway into the file system.

###DESCRIPTION###
<p>The FileView class is designed for developing graphical interfaces that
require user interactive file lists.  This makes the FileView particularly
useful for creating file dialog boxes and file managers.  The class supports
a number of features including the filtering of file names using wildcards
(for example, '*.dml') and you may preset the initial directory location.</p>

<p>The FileView class does not draw its own graphics.  In order to
display the file-list, you need to provide it with a data view object by
setting the <a href="#View">View</a> field.  Currently we recommend that you
use a View object to provide viewing services.  The following DML
example shows how to construct an effective file list that loads a text
editor when the user double clicks on a file:</p>

<pre>
  &lt;render name="main" x="5" y="5" xoffset="5" colour="#f0f0f0" yoffset="5"&gt;
    &lt;scrollbar direction="vertical"/&gt;

    &lt;view style="list" x="5" y="4" linespacing="4" highlight="#d0d0e0"&gt;
      &lt;fileview view="[owner]" location="athene:" filter="*.dml"&gt;
        &lt;create class="run" execute &amp;src="bin:runscript"
          &amp;args="tools:text/editor/main.dml -args location='\[owner.selectionfile\]'"/&gt;
        &lt;action static call="free" object="[main]"/&gt;
      &lt;/fileview&gt;
    &lt;/text&gt;
  &lt;/render&gt;
</pre>

<p>It is possible to create a refined FileView interface by linking buttons
and input boxes directly to the object.  For a working example, please refer
to the 'templates:filedialog.dml' file and observe how we created the default
file requester for Athene.</p>

###END###

*****************************************************************************/

//#define DEBUG

#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/files/all.h>
#include <pandora/misc/all.h>
#include <pandora/modules/strings.h>
#include <pandora/modules/filesystem.h>
#include <pandora/input/joyport.h>
#include <pandora/input/keyboard.h>
#include <pandora/input/clipboard.h>
#include <pandora/tools/dialog.h>
#include <pandora/tools/fileview.h>
#include <pandora/data/xml.h>
#include <pandora/data/config.h>
#include <pandora/functions/set.c>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_FILEVIEW, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2004. All rights reserved.",
   "March 2004",
   "FileView"
};

#define TITLE_RENAME    "Rename"
#define TITLE_CREATEDIR "Create New Directory"
#define TITLE_DELETE    "Confirm Deletion"

static struct KernelBase     *KernelBase;
static struct FileSystemBase *FileSystemBase;
static struct StringsBase    *StringsBase;
static OBJECTPTR FileViewClass = NULL, glClipboard = NULL;
static OBJECTPTR modStrings = NULL, modFileSystem = NULL;
static FIELD FID_NewName;

/*****************************************************************************
** Object structure.
*/

struct FileView {
   OBJECT_HEADER         /* Class header */
   OBJECTID ViewID;      /* External data view */
   LONG     Flags;       /* Optional flags */
   FLOAT    RefreshRate; /* Refresh the file-list every X number of seconds */
   OBJECTID InputID;     /* The keyboard to read information from */
   OBJECTID FocusID;     /* The drawable to monitor for the focus */

   PRIVATE_FIELDS
   UBYTE  Filter[40];  /* File filter */
   LONG   FileCount;
   UBYTE  Selection[130];     /* Currently selected file/directory */
   UBYTE  Location[160];
   BYTE   Refresh;
   BYTE   KeyboardChannel;    /* Set to true if keyboard channel is active */
   STRING SelectionPath;      /* The complete, fully qualified file-path of the current selection */
   struct DirInfo *Dir;
   LONG   DeleteIndex;
   BYTE   ResetTimer;
};

/*****************************************************************************
** Class definition.
*/

static ERROR GET_Filter(struct FileView *, STRING *);
static ERROR GET_Location(struct FileView *, STRING *);
static ERROR GET_Selection(struct FileView *, STRING *);
static ERROR GET_SelectionFile(struct FileView *, STRING *);
static ERROR GET_SelectionPath(struct FileView *, STRING *);

static ERROR SET_Filter(struct FileView *, STRING);
static ERROR SET_Location(struct FileView *, STRING);

static struct FieldDef ViewFlags[] = {
   { "SHOWPARENT",   FVF_SHOWPARENT },
   { "NOEXTENSIONS", FVF_NOEXTENSIONS },
   { "SYSKEYS",      FVF_SYSKEYS },
   { "NOFILES",      FVF_NOFILES },
   { NULL, NULL }
};

static struct FieldArray FileViewFields[] = {
   { "View",        0, FDF_OBJECTID|FDF_RI,  0,           NULL, NULL },
   { "Flags",       0, FDF_LONGFLAGS|FDF_RW, (LONG)&ViewFlags, NULL, NULL },
   { "RefreshRate", 0, FDF_FLOAT|FDF_RW,     0,           NULL, NULL },
   { "Input",       0, FDF_OBJECTID|FDF_RI,  ID_KEYBOARD, NULL, NULL },
   { "Focus",       0, FDF_OBJECTID|FDF_RI,  0,           NULL, NULL },
   /** Virtual fields ***/
   { "Filter",        0, FDF_STRING|FDF_RW, 0, GET_Filter,    SET_Filter },
   { "Location",      0, FDF_STRING|FDF_RW, 0, GET_Location,  SET_Location },
   { "Selection",     0, FDF_STRING|FDF_R,  0, GET_Selection, NULL },
   { "SelectionFile", 0, FDF_STRING|FDF_R,  0, GET_SelectionFile, NULL },
   { "SelectionPath", 0, FDF_STRING|FDF_R,  0, GET_SelectionPath, NULL },
   END_FIELD
};

/****************************************************************************/

static ERROR FILEVIEW_ActionNotify(struct FileView *, struct acActionNotify *);
static ERROR FILEVIEW_Activate(struct FileView *, APTR);
static ERROR FILEVIEW_DataChannel(struct FileView *, struct acDataChannel *);
static ERROR FILEVIEW_FieldNotify(struct FileView *, struct acFieldNotify *);
static ERROR FILEVIEW_Focus(struct FileView *, APTR);
static ERROR FILEVIEW_Free(struct FileView *, APTR);
static ERROR FILEVIEW_Init(struct FileView *, APTR);
static ERROR FILEVIEW_LostFocus(struct FileView *, APTR);
static ERROR FILEVIEW_NewObject(struct FileView *, APTR);
static ERROR FILEVIEW_Refresh(struct FileView *, APTR);
static ERROR FILEVIEW_Timer(struct FileView *, struct acTimer *);

static struct ActionArray FileViewActions[] = {
   { AC_ActionNotify,     FILEVIEW_ActionNotify },
   { AC_Activate,         FILEVIEW_Activate },
   { AC_DataChannel,      FILEVIEW_DataChannel },
   { AC_FieldNotify,      FILEVIEW_FieldNotify },
   { AC_Focus,            FILEVIEW_Focus },
   { AC_Free,             FILEVIEW_Free },
   { AC_Init,             FILEVIEW_Init },
   { AC_LostFocus,        FILEVIEW_LostFocus },
   { AC_NewObject,        FILEVIEW_NewObject },
   { AC_Refresh,          FILEVIEW_Refresh },
   { AC_Timer,            FILEVIEW_Timer },
   { NULL, NULL }
};

/****************************************************************************/

static ERROR FILEVIEW_CutFiles(struct FileView *, APTR);
static ERROR FILEVIEW_CopyFiles(struct FileView *, APTR);
static ERROR FILEVIEW_CopyFilesTo(struct FileView *, struct mtCopyFilesTo *);
static ERROR FILEVIEW_CreateDir(struct FileView *, APTR);
static ERROR FILEVIEW_DeleteFiles(struct FileView *, APTR);
static ERROR FILEVIEW_EditFiles(struct FileView *, APTR);
static ERROR FILEVIEW_MoveFilesTo(struct FileView *, struct mtMoveFilesTo *);
static ERROR FILEVIEW_ParentDirectory(struct FileView *, APTR);
static ERROR FILEVIEW_PasteFiles(struct FileView *, APTR);
static ERROR FILEVIEW_RenameFile(struct FileView *, APTR);
static ERROR FILEVIEW_ViewFiles(struct FileView *, APTR);

static struct FunctionField argsCopyFilesTo[] = { { "Dest", ARG_STR }, { NULL, NULL } };
static struct FunctionField argsMoveFilesTo[] = { { "Dest", ARG_STR }, { NULL, NULL } };

static struct MethodArray FileViewMethods[] = {
   { MT_CutFiles,        FILEVIEW_CutFiles,        "CutFiles",        NULL, NULL },
   { MT_CopyFiles,       FILEVIEW_CopyFiles,       "CopyFiles",       NULL, NULL },
   { MT_CopyFilesTo,     FILEVIEW_CopyFilesTo,     "CopyFilesTo",     argsCopyFilesTo, sizeof(struct mtCopyFilesTo) },
   { MT_CreateDir,       FILEVIEW_CreateDir,       "CreateDir",       NULL, NULL },
   { MT_DeleteFiles,     FILEVIEW_DeleteFiles,     "DeleteFiles",     NULL, NULL },
   { MT_MoveFilesTo,     FILEVIEW_MoveFilesTo,     "MoveFilesTo",     argsMoveFilesTo, sizeof(struct mtMoveFilesTo) },
   { MT_ParentDirectory, FILEVIEW_ParentDirectory, "ParentDirectory", NULL, NULL },
   { MT_PasteFiles,      FILEVIEW_PasteFiles,      "PasteFiles",      NULL, NULL },
   { MT_RenameFile,      FILEVIEW_RenameFile,      "RenameFile",      NULL, NULL },
   { MT_EditFiles,       FILEVIEW_EditFiles,       "EditFiles",       NULL, NULL },
   { MT_ViewFiles,       FILEVIEW_ViewFiles,       "ViewFiles",       NULL, NULL },
   { NULL, NULL, NULL, NULL }
};

/*****************************************************************************
** Function prototypes.
*/

static void InitKeyboard(struct FileView *);
static void AddFileItem(struct FileView *, OBJECTPTR, struct DirInfo *);
static ERROR RenameFileItem(struct FileView *, STRING, STRING);
static void StripExtension(STRING);
static STRING ExtractFilename(struct XMLTag *);
static ERROR PasteTo(struct FileView *, STRING);
static void convert_permissions(LONG, STRING);

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   /*** Open required modules ***/

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   if (CreateObject(ID_MODULE, NULL, &modFileSystem, NULL,
      FID_Name|TSTRING,   "filesystem",
      FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
      TAGEND) IS ERR_Okay) {
      if (GetField(modFileSystem, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   ResolveFields("NewName", &FID_NewName, TAGEND);

   if (CreateObject(ID_CLIPBOARD, NULL, &glClipboard, NULL, TAGEND) != ERR_Okay) {
      return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));
   }

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&FileViewClass, NULL,
      FID_BaseClassID|TLONG, ID_FILEVIEW,
      FID_Version|TFLOAT,    VER_FILEVIEW,
      FID_Name|TSTRING,      "FileView",
      FID_Category|TLONG,    CCF_TOOL,
      FID_Actions|TPTR,      FileViewActions,
      FID_Methods|TPTR,      FileViewMethods,
      FID_Fields|TPTR,       FileViewFields,
      FID_Size|TLONG,        sizeof(struct FileView),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (glClipboard)   { acFree(glClipboard);   glClipboard = NULL; }
   if (FileViewClass) { acFree(FileViewClass); FileViewClass = NULL; }
   if (modStrings)    { acFree(modStrings);    modStrings = NULL; }
   if (modFileSystem) { acFree(modFileSystem); modFileSystem = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** FileView: ActionNotify
**
** NOTE: See FILEVIEW_FieldNotify for other notification response routines.
*/

static ERROR FILEVIEW_ActionNotify(struct FileView *Self, struct acActionNotify *Args)
{
   if (Args->Error != ERR_Okay) return(ERR_Okay);

   if (Args->ActionID IS AC_Focus) {
      InitKeyboard(Self);
   }
   else if (Args->ActionID IS AC_LostFocus) {
      acLostFocus(Self);
   }

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Activate
Short: Opens the currently selected file or directory.

###DESCRIPTION###
<p>If the user has selected a file or a directory, calling this method will
'open' the selection accordingly.  For directories, this means that the
FileView object will go to the selected location and read the directory
contents, consequently updating the FileView.  If the selection is a file,
the FileView will send an Activate message to each child that you have
initialised to the FileView object.</p>

<p>If no selection has been made in the <a href="#View">View</a> object,
the Activate action will do nothing.</p>

###RESULT###
ERR_Okay:            The selection was successfully opened.
ERR_AllocMemory:     A call to AllocMemory() failed.
ERR_ExclusiveDenied: Access to the View object was denied.
ERR_GetField:        The View does not support a Selection field, or no selection has been made by the user.
ERR_ListChildren:    A call to ListChildren() failed.

###END###

*****************************************************************************/

static ERROR FILEVIEW_Activate(struct FileView *Self, APTR Void)
{
   struct XMLTag **taglist;
   OBJECTPTR xml;
   OBJECTPTR view;
   STRING copy, str, name;
   LONG size, i, count, tagindex;
   ERROR error;

   /* Note: Activate notification is silent when the
   ** user is simply switching directories.
   */

   DPrintF("~Activate()","[FileView:%d]", Self->Head.UniqueID);

   if (AccessObject(Self->ViewID, 5000, &view) IS ERR_Okay) {
      if ((GetFields(view, FID_XML|TPTR,        &xml,
                           FID_ActiveTag|TLONG, &tagindex,
                           FID_Tags|TPTR,       &taglist,
                           TAGEND) IS ERR_Okay) AND (tagindex != -1) AND (xml)) {

         /*** Check if the selected item is a directory ***/

         name = ExtractFilename(taglist[tagindex]);
         for (i=0; (name[i]) AND (name[i] != ':') AND (name[i] != '/'); i++);

         if ((name[i] IS ':') OR (name[i] IS '/')) {
            /*** Respond by switching to the new directory ***/
            size = StrLength(Self->Location) + StrLength(name) + 2;
            if (AllocMemory(size, MEM_STRING, (APTR *)&str, NULL) IS ERR_Okay) {
               copy = str;
               if ((Self->Location) AND (Self->Location[0] != ':')) copy += StrCopy(Self->Location, copy, COPY_ALL);
               copy += StrCopy(name, copy, COPY_ALL);
               if ((copy[-1] != ':') AND (copy[-1] != '/')) {
                  *copy++ = '/';
                  *copy++ = 0;
               }
               SetField(Self, FID_Location, FT_STRING, str);
               FreeMemory(str);
               error = ERR_Okay|ERF_Notified;
            }
            else error = ObjectError(Self, ERH_Activate, ERR_AllocMemory);
         }
         else if ((Self->Flags & FVF_SHOWPARENT) AND (Self->Location[0] != ':') AND (tagindex IS 0)) {
            /*** Go to the parent directory ***/
            error = FILEVIEW_ParentDirectory(Self, NULL) | ERF_Notified;
         }
         else if ((count = TotalChildren(Self->Head.UniqueID))) {
            struct ChildEntry list[count];
            if (ListChildren(GetUniqueID(Self), list, &count) IS ERR_Okay) {
               for (i=0; i < count; i++) ActionMsg(AC_Activate, list[i].ObjectID, NULL);
               error = ERR_Okay;
            }
            else error = ERR_ListChildren;
         }
         else error = ERR_Okay;
      }
      else error = ERR_GetField;
      ReleaseObject(view);
   }
   else error = ObjectError(Self, ERH_Activate, ERR_ExclusiveDenied);

   StepBack();
   return(error);
}

/*****************************************************************************

###METHOD###
Name:  CopyFiles
Short: Posts all currently selected files to the clipboard as a 'copy' operation.

###DESCRIPTION###
<p>The CopyFiles method will copy all user-selected files and directories from
the view to the clipboard.  If there are no files selected, the method does
nothing.</p>

###RESULT###
ERR_Okay: The cut operation was processed.
ERR_NothingDone: The user has not selected any files to copy.
ERR_AccessObject: Failed to access either the clipboard or the view object.
ERR_NoSupport: No clipboard is present in the system to support the copy operation.

###END###

*****************************************************************************/

static ERROR FILEVIEW_CopyFiles(struct FileView *Self, APTR Void)
{
   struct XMLTag **taglist;
   OBJECTPTR view, xml;
   LONG i, j, *tags;
   UBYTE buffer[512];
   ERROR error;

   /*** Do nothing if we are at the root level ***/

   if ((!Self->Location[0]) OR (Self->Location[0] IS ':')) return(ERR_Okay);

   DPrintF("~CopyFiles()","[FileView:%d]", Self->Head.UniqueID);

   if (AccessObject(Self->ViewID, 5000, &view) IS ERR_Okay) {
      if ((GetFields(view, FID_SelectedTags|TPTR, &tags,
                           FID_XML|TPTR,          &xml,
                           FID_Tags|TPTR,         &taglist,
                           TAGEND) IS ERR_Okay) AND (tags)) {

         /*** Clear any existing file clip records ***/

         ActionTags(MT_DeleteClip, glClipboard, CLIP_FILE);

         /*** Add all selected files to the clipboard ***/

         j = StrCopy(Self->Location, buffer, sizeof(buffer));
         for (i=0; tags[i] != -1; i++) {
            StrCopy(ExtractFilename(taglist[tags[i]]), buffer+j, sizeof(buffer)-j);
            ActionTags(MT_AddClipboardFiles, glClipboard, CLIP_FILE, buffer, CLF_EXTEND);
         }

         error = ERR_Okay;
      }
      else error = ERR_NothingDone;

      ReleaseObject(view);
   }
   else {
      DPrintF("@CopyFiles:","Failed to access view.");
      error = ERR_AccessObject;
   }

   StepBack();
   return(error);
}

/*****************************************************************************

###METHOD###
Name:  CopyFilesTo
Short: Copies selected files to a destination location.

###DESCRIPTION###
<p>This method copies all selected files in the fileview to a destination
directory of your choosing.  If no files are selected in the fileview, this
method does nothing.  The CopyFilesTo method may return immediately following
the initial call and then copy the files in the background.  If a failure
occurs, the user will be notified with an error dialog.</p>

###INPUT###
Dest: The directory that you want to copy the selected files to.

###END###

*****************************************************************************/

static ERROR FILEVIEW_CopyFilesTo(struct FileView *Self, struct mtCopyFilesTo *Args)
{
   ERROR error;

   if ((!Args) OR (!Args->Dest) OR (!Args->Dest[0])) return(ObjectError(Self, MT_CopyFilesTo, ERR_Args));

   DPrintF("~CopyFilesTo()","[FileView:%d]", Self->Head.UniqueID);

   if ((error = FILEVIEW_CopyFiles(Self, NULL)) IS ERR_Okay) {
      error = PasteTo(Self, Args->Dest);
   }

   ActionMsg(MT_SelectNone, Self->ViewID, NULL);

   StepBack();

   return(error);
}

/*****************************************************************************

###METHOD###
Name:  CreateDir
Short: Creates a new directory in the current path of the file view.

###DESCRIPTION###
<p>This method creates a dialog box that prompts the user for a directory name.
If the user types in a valid directory name then the directory will be created
in the current location of the file view.  The user may cancel the process
by closing the dialog window at any time.</p>

###END###

*****************************************************************************/

static ERROR FILEVIEW_CreateDir(struct FileView *Self, APTR Void)
{
   extern UBYTE glNewShortcutScript[];
   OBJECTID dialogid;
   OBJECTPTR script, dialog;
   ERROR error;

   DPrintF("~CreateDir()","[FileView:%d] Location: %s", Self->Head.UniqueID, Self->Location);

   if ((!Self->Location[0]) OR (Self->Location[0] IS ':')) {
      /*** Create a new shortcut/assignment at the root level ***/

      if (CreateObject(ID_SCRIPT, NF_CHILD, &script, NULL,
         FID_String|TSTRING, glNewShortcutScript,
         TAGEND) IS ERR_Okay) {
         error = acActivate(script);
         acFree(script);
      }
      else error = ERR_CreateObject;
   }
   else {
      if (CreateObject(ID_DIALOG, NF_CHILD, &dialog, &dialogid,
         FID_Image|TSTRING,    "icons:folders/folder_new(48)",
         FID_Buttons|TSTRING,  "okay;cancel",
         FID_Title|TSTRING,    TITLE_CREATEDIR,
         FID_Flags|TLONG,      DF_INPUT|DF_INPUTREQUIRED,
         FID_String|TSTRING,   "Please enter the name of the new directory that you wish to create.",
         FID_Action|TUNLISTED, "CreateDir",
         TAGEND) IS ERR_Okay) {

         SetUnlistedField(dialog, "Dir", Self->Location);
         SubscribeField(dialog, FID_Response, Self);
         acShow(dialog);
         error = ERR_Okay;
         ReleaseObject(dialog);
      }
      else error = ERR_CreateObject;
   }

   StepBack();
   return(error);
}

/*****************************************************************************

###METHOD###
Name:  CutFiles
Short: Posts all currently selected files to the clipboard as a 'cut' operation.

###DESCRIPTION###
<p>The CutFiles method will post all selected files and directories from
the view to the clipboard.  If there are no files selected, the method does
nothing.</p>

###END###

*****************************************************************************/

static ERROR FILEVIEW_CutFiles(struct FileView *Self, APTR Void)
{
   struct XMLTag **taglist;
   OBJECTPTR view, xml;
   LONG i, j, *tags;
   UBYTE buffer[300];

   /*** Do nothing if we are at the root level ***/

   if ((!Self->Location[0]) OR (Self->Location[0] IS ':')) return(ERR_Okay);

   DPrintF("~CutFiles()","[FileView:%d]", Self->Head.UniqueID);

   if (AccessObject(Self->ViewID, 5000, &view) IS ERR_Okay) {
      if ((GetFields(view, FID_SelectedTags|TPTR, &tags,
                           FID_XML|TPTR,          &xml,
                           FID_Tags|TPTR,         &taglist,
                           TAGEND) IS ERR_Okay) AND (tags)) {

         /*** Clear any existing file clips ***/

         ActionTags(MT_DeleteClip, glClipboard, CLIP_FILE);

         /*** Add all selected files to the clipboard ***/

         j = StrCopy(Self->Location, buffer, sizeof(buffer));
         for (i=0; tags[i] != -1; i++) {
            StrCopy(ExtractFilename(taglist[tags[i]]), buffer+j, sizeof(buffer)-j);
            ActionTags(MT_AddClipboardFiles, glClipboard, CLIP_FILE, buffer, CLF_EXTEND|CLF_DELETE);
         }
      }

      ReleaseObject(view);
   }
   else DPrintF("@CutFiles:","Failed to access view.");

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** FileView: DataChannel
*/

static ERROR FILEVIEW_DataChannel(struct FileView *Self, struct acDataChannel *Args)
{
   LONG keyvalue, keyflags, j;

   /* Pass XML information on the XML object.  Then refresh the display to
   ** include the new information.
   */

   if (!Args) return(ObjectError(Self, ERH_DataChannel, ERR_Args));

   if ((Args->DataType IS DATA_KEYPRESS) AND (Args->Version <= 1)) {

      for (j=0; j < Args->TotalEntries; j++) {
         if (((struct dcKeyEntry *)Args->Buffer)[j].Flags & KQ_PRESSED) {
            keyflags = ((struct dcKeyEntry *)Args->Buffer)[j].Flags;
            keyvalue = ((struct dcKeyEntry *)Args->Buffer)[j].Value;

            if ((Self->Flags & FVF_SYSKEYS) AND (keyflags & KQ_CTRL)) {
               /* Note: Syskeys should only be enabled if cut/copy/paste
               ** keyboard support is not implemented elsewhere, such as menu items.
               */
               switch(keyvalue) {
                  case K_C: /* copy */
                     Action(MT_CopyFiles, Self, NULL);
                     break;

                  case K_X: /* cut */
                     Action(MT_CutFiles, Self, NULL);
                     break;

                  case K_V: /* paste */
                     Action(MT_PasteFiles, Self, NULL);
                     break;
               }
            }
            else switch(keyvalue) {
                case K_DELETE:
                   Action(MT_DeleteFiles, Self, NULL);
                   break;
            }
         }
      }
      return(ERR_Okay);
   }
   else DPrintF("@DataChannel:","[FileView:%d] Data type #%d not supported.", Self->Head.UniqueID, Args->DataType);

   return(ERR_NoSupport);
}

/*****************************************************************************

###METHOD###
Name:  DeleteFiles
Short: Deletes all currently selected files, following user confirmation.

###DESCRIPTION###
<p>This method simplifies the process of deleting selected files for the user.  The
method will pop-up a dialog box to ask the user if the selected files should
be deleted.  If the user responds positively, the method will proceed to delete
all of the requested files.  Error dialogs may pop-up if any problems occur
during the deletion process.</p>

<p>If no files are selected, this method does nothing.</p>

<p>This method will return a failure code if the creation of the initial
dialog box fails.</p>

###RESULT###
ERR_Okay: The process executed successfully.
ERR_CreateObject: Failed to create the dialog box.
ERR_AccessObject: Failed to access the referenced View object.

###END###

*****************************************************************************/

static ERROR FILEVIEW_DeleteFiles(struct FileView *Self, APTR Void)
{
   OBJECTPTR dialog, view, xml, config;
   struct XMLTag **taglist;
   LONG *tags, i;
   ERROR error;
   UBYTE buffer[512], intstr[20];

   DPrintF("~DeleteFiles()","[FileView:%d] Location: %s", Self->Head.UniqueID, Self->Location);

   if (AccessObject(Self->ViewID, 5000, &view) IS ERR_Okay) {
      if ((GetFields(view, FID_SelectedTags|TPTR, &tags,
                           FID_XML|TPTR,          &xml,
                           FID_Tags|TPTR,         &taglist,
                           TAGEND) IS ERR_Okay) AND (tags)) {

         for (i=0; tags[i] != -1; i++);

         if (i > 1) StrFormat(buffer, sizeof(buffer), "Are you sure that you wish to delete the %d selected items?", i);
         else {
             i = StrCopy("Are you sure that you wish to delete the '", buffer, sizeof(buffer));

             i += StrCopy(ExtractFilename(taglist[tags[0]]), buffer+i, sizeof(buffer)-i);

             if (buffer[i-1] IS '/') StrCopy("' folder?", buffer+i-1, sizeof(buffer)-i);
             else if (buffer[i-1] IS ':') StrCopy("' shortcut?", buffer+i-1, sizeof(buffer)-i);
             else StrCopy("' file?", buffer+i, sizeof(buffer)-i);
         }

         if (CreateObject(ID_DIALOG, NF_CHILD, &dialog, NULL,
            FID_Image|TSTRING,    "icons:tools/eraser(48)",
            FID_Buttons|TSTRING,  "YESALL:Yes; CANCEL:No",
            FID_Title|TSTRING,    TITLE_DELETE,
            FID_String|TSTRING,   buffer,
            FID_Action|TUNLISTED, "Delete",
            TAGEND) IS ERR_Okay) {

            /* Create a Config object that will store all of the
            ** files that we are going to delete.  When the user
            ** responds to the dialog box positively, we'll use the
            ** config object's content to determine what we're going
            ** to delete (see the FieldNotify support action).
            */

            if (CreateObject(ID_CONFIG, NULL, &config, NULL,
               FID_Owner|TLONG, dialog->UniqueID,
               TAGEND) IS ERR_Okay) {

               for (i=0; tags[i] != -1; i++) {
                  StrFormat(buffer, sizeof(buffer), "%s%s", Self->Location, ExtractFilename(taglist[tags[i]]), buffer);
                  IntToStr(i, intstr, sizeof(intstr));
                  ActionTags(MT_WriteConfig, config, "DELETEFILES", intstr, buffer);
               }
            }

            SubscribeField(dialog, FID_Response, Self);

            acShow(dialog);

            error = ERR_Okay;
         }
         else error = ERR_CreateObject;
      }
      else error = ERR_Okay;

      ReleaseObject(view);
   }
   else { DPrintF("@DeleteFiles:","Failed to access view."); error = ERR_AccessObject; }

   StepBack();
   return(error);
}

/*****************************************************************************

###METHOD###
Name:  EditFiles
Short: Runs the edit command for all currently selected files.

###DESCRIPTION###
<p>This method simplifies the process of editing selected files for the user.
The routine utilises the Run class with the EDIT mode option to launch the files
in their respected editors.</p>

<p>If no files are selected, this method does nothing.</p>

###RESULT###
ERR_Okay: The process executed successfully.
ERR_AccessObject: Failed to access the referenced View object.

###END###

*****************************************************************************/

static ERROR FILEVIEW_EditFiles(struct FileView *Self, APTR Void)
{
   OBJECTPTR run, xml, view;
   struct XMLTag **taglist;
   LONG *tags;
   UBYTE buffer[300];
   LONG i;

   DPrintF("~EditFiles()","[FileView:%d]", Self->Head.UniqueID);

   if ((!Self->Location[0]) OR (Self->Location[0] IS ':')) {
      StepBack();
      return(ERR_Okay);
   }

   if (AccessObject(Self->ViewID, 5000, &view) IS ERR_Okay) {
      if ((GetFields(view, FID_SelectedTags|TPTR, &tags,
                           FID_XML|TPTR,          &xml,
                           FID_Tags|TPTR,         &taglist,
                           TAGEND) IS ERR_Okay) AND (tags)) {

         for (i=0; tags[i] != -1; i++) {
            StrFormat(buffer, sizeof(buffer), "%s%s", Self->Location, ExtractFilename(taglist[tags[i]]));

            if (CreateObject(ID_RUN, NULL, &run, NULL,
               FID_Mode|TSTRING,     "edit",
               FID_Location|TSTRING, buffer,
               TAGEND) IS ERR_Okay) {
               acActivate(run);
               acFree(run);
            }
         }
      }
      ReleaseObject(view);
   }
   else return(StepObjectError(Self, MT_EditFiles, ERR_AccessObject));

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************
** FileView: FieldNotify
*/

static ERROR FILEVIEW_FieldNotify(struct FileView *Self, struct acFieldNotify *Args)
{
   struct ChildEntry list[5];
   UBYTE buffer[200];
   struct ConfigEntry *entries;
   OBJECTPTR dialog, config, errordialog, view, xml;
   OBJECTID configid;
   STRING errorfile, errorstr;
   LONG response, count, amtentries, i, fstart, errorcount;

   if (!Args) return(ObjectError(Self, ERH_FieldNotify, ERR_Args));

   DPrintF("~FieldNotify:","[FileView] Response notification received.");

   if (AccessObject(Args->ObjectID, 5000, &dialog) IS ERR_Okay) {
      GetUnlistedField(dialog, "Action", buffer, sizeof(buffer));

      if (StrCompare("Rename", buffer, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         GetField(dialog, FID_Response, FT_LONG, &response);

         if (response & RSF_POSITIVE) {
            STRING newname;
            OBJECTPTR file, confirmdialog;
            UBYTE src[1024], dest[1024], dir, state;
            ERROR error;

            errorstr = NULL;

            if (GetUnlistedField(dialog, "Dest", dest, sizeof(dest)) IS ERR_Okay) {
               if (GetUnlistedField(dialog, "Src", src, sizeof(src)) IS ERR_Okay) {
                  if (MoveLocation(src, dest) IS ERR_Okay) {
                     DelayMsg(AC_Refresh, Self->Head.UniqueID, NULL);
                  }
                  else errorstr = "The rename operation failed.";
               }
            }
            else {
               if (GetUnlistedField(dialog, "Src", src, sizeof(src)) != ERR_Okay) src[0] = 0;
               GetField(dialog, FID_TextInput, FT_STRING, &newname);

               i = StrCopy(src, dest, sizeof(dest));
               if ((dest[i-1] IS '/') OR (dest[i-1] IS ':') OR (dest[i-1] IS '\\')) {
                  i--; /* The source being renamed is a directory */
                  dir = TRUE;
               }
               else dir = FALSE;

               while ((i > 0) AND (dest[i-1] != '/') AND (dest[i-1] != '\\') AND (dest[i-1] != ':')) i--;
               fstart = i;
               i += StrCopy(newname, dest+i, sizeof(dest)-i-1);
               if (dir) {
                  dest[i++] = '/';
                  dest[i] = 0;
               }

               if (StrCompare(src, dest, 0, STR_MATCHLENGTH|STR_MATCHCASE) != ERR_Okay) {
                  state = FileCaseSensitive(TRUE);
                  error = AnalyseLocation(dest, NULL);
                  FileCaseSensitive(state);

                  if (error IS ERR_Okay) {
                     /*** The destination exists ***/

                     if (dir) dest[i-1] = 0;

                     if (CreateObject(ID_DIALOG, NF_CHILD, &confirmdialog, NULL,
                        FID_Image|TSTRING,       "icons:items/question(48)",
                        FID_Buttons|TSTRING,     "YES:Yes; CANCEL:No",
                        FID_Title|TSTRING,       "Confirmation Required",
                        FID_String|TSTRING,      "A file already exists that uses this name.  Are you sure that you want to overwrite it?",
                        FID_Action|TUNLISTED,    "Rename",
                        FID_Src|TUNLISTED,       src,
                        FID_Dest|TUNLISTED,      dest,
                        TAGEND) IS ERR_Okay) {
                        SubscribeField(confirmdialog, FID_Response, Self);
                        acShow(confirmdialog);
                     }
                  }
                  else if (CreateObject(ID_FILE, NF_CHILD, &file, NULL,
                     FID_Location|TSTRING, src,
                     FID_Flags|TLONG,      FL_READ,
                     TAGEND) IS ERR_Okay) {
                     if (ActionTags(AC_Rename, file, newname) IS ERR_Okay) {
                        RenameFileItem(Self, src+fstart, newname);
                     }
                     else errorstr = "Failed to rename the file.";

                     acFree(file);
                  }
                  else errorstr = "Failed to open file for renaming.";
               }
            }

            if (errorstr) {
               if (CreateObject(ID_DIALOG, NF_CHILD, &errordialog, NULL,
                  FID_Image|TSTRING,   "icons:items/error(48)",
                  FID_Buttons|TSTRING, "NONE:Okay",
                  FID_Title|TSTRING,   "File Rename Failure",
                  FID_String|TSTRING,  errorstr,
                  TAGEND) IS ERR_Okay) {

                  acShow(errordialog);
               }
            }
         }
      }
      else if (StrCompare("CreateDir", buffer, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         struct FileInfo info;
         UBYTE dir[1024];
         STRING name;
         GetField(dialog, FID_Response, FT_LONG, &response);

         if (response & RSF_POSITIVE) {
            GetUnlistedField(dialog, "Dir", dir, sizeof(dir));
            for (i=0; dir[i]; i++);
            if ((GetField(dialog, FID_TextInput, FT_STRING, &name) IS ERR_Okay) AND (name) AND (name[0])) {
               i += StrCopy(name, dir+i, sizeof(dir)-i);
               dir[i++] = '/';
               dir[i] = 0;
               if (MakeDirectory(dir, NULL) IS ERR_Okay) {
                  if (AccessObject(Self->ViewID, 5000, &view) IS ERR_Okay) {
                     if ((GetField(view, FID_XML, FT_POINTER, &xml) IS ERR_Okay) AND (xml)) {
                        if (GetFileInfo(dir, &info, sizeof(info)) IS ERR_Okay) {
                           AddFileItem(Self, xml, (struct DirInfo *)&info);
                           acSort(view);
                           acRefresh(view);
                        }
                     }
                     ReleaseObject(view);
                  }
               }
               else DPrintF("@FileView:","Failed to create dir \"%s\"", dir);
            }
            else DPrintF("@FileView:","No name provided for dir creation.");
         }
      }
      else if (StrCompare("Delete", buffer, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         GetField(dialog, FID_Response, FT_LONG, &response);

         if (response IS RSF_YESALL) {
            count = ARRAYSIZE(list);
            ListChildren(dialog->UniqueID, list, &count);

            configid = NULL;
            for (i=0; i < ARRAYSIZE(list); i++) {
               if (list[i].ClassID IS ID_CONFIG) {
                  configid = list[i].ObjectID;
                  break;
               }
            }

            if ((configid) AND (AccessObject(configid, 5000, &config) IS ERR_Okay)) {
               if (GetFields(config, FID_Entries|TPTR,     &entries,
                                     FID_AmtEntries|TLONG, &amtentries,
                                     TAGEND) IS ERR_Okay) {

                  errorcount = 0;
                  errorfile  = NULL;
                  for (i=0; i < amtentries; i++) {
                     if (DeleteFile(entries[i].Data) != ERR_Okay) {
                        errorfile = entries[i].Data;
                        errorcount++;
                     }
                  }

                  if (errorcount > 0) {
                     DPrintF("FieldNotify:","Errors detected during delete process (%d files)", errorcount);

                     if (errorcount IS 1) StrFormat(buffer, sizeof(buffer), "The following file could not be deleted due to an error:\n\n%s", errorfile);
                     else StrFormat(buffer, sizeof(buffer), "%d files could not be deleted due to errors.", errorcount);

                     if (CreateObject(ID_DIALOG, NF_CHILD, &errordialog, NULL,
                           FID_Image|TSTRING,   "icons:items/error(48)",
                           FID_Buttons|TSTRING, "NONE:Okay",
                           FID_Title|TSTRING,   "File Deletion Failures",
                           FID_String|TSTRING,  buffer,
                           TAGEND) IS ERR_Okay) {

                        acShow(errordialog);
                     }
                  }

                  acRefresh(Self);
               }

               ReleaseObject(config);
            }
         }
      }

      ReleaseObject(dialog);
   }

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  Focus
Short: Activates keyboard monitoring (by force).
###END###

*****************************************************************************/

static ERROR FILEVIEW_Focus(struct FileView *Self, APTR Void)
{
   return(acFocusID(Self->ViewID));
}

/*****************************************************************************
** FileView: Free
*/

static ERROR FILEVIEW_Free(struct FileView *Self, APTR Void)
{
   OBJECTPTR object;

   /*** Release the keyboard ***/

   if ((Self->InputID) AND (Self->KeyboardChannel IS TRUE)) {
      if (AccessObject(Self->InputID, 5000, &object) IS ERR_Okay) {
         UnsubscribeChannel(object, Self->Head.UniqueID);
         ReleaseObject(object);
      }
      Self->InputID = NULL;
   }

   if (Self->SelectionPath) { FreeMemory(Self->SelectionPath); Self->SelectionPath = NULL; }

   if (Self->Dir) { CloseDirectory(Self->Dir); Self->Dir = NULL; }

   return(ERR_Okay);
}

/*****************************************************************************
** FileView: Init()
*/

static ERROR FILEVIEW_Init(struct FileView *Self, APTR Void)
{
   OBJECTPTR view, object;

   if (GetClassID(Self->ViewID) != ID_VIEW) {
      DPrintF("@FileView:","The View field must be set to refer to a View object.");
      return(ERR_Failed);
   }

   if (!Self->FocusID) {
      if (AccessObject(Self->ViewID, 5000, &view) IS ERR_Okay) {
         GetField(view, FID_Drawable, FT_LONG, &Self->FocusID);
         ReleaseObject(view);
      }
   }

   if (AccessObject(Self->FocusID, 5000, &object) IS ERR_Okay) {
      SubscribeActionTags(object, Self,
         AC_Focus,
         AC_LostFocus,
         TAGEND);

      /* Test the focus object to see if it already has the focus and
      ** if so, activate keyboard monitoring.
      */

      if (object->ClassID IS ID_RENDER) {
         if (((struct Render *)object)->Flags & RNF_HASFOCUS) InitKeyboard(Self);
      }

      ReleaseObject(object);
   }

   /*** Setup keyboard ***/

   if (!Self->InputID) {
      if (FastFindObject("SystemKeyboard", NULL, &Self->InputID, 1, NULL) != ERR_Okay) {
         DPrintF("@Init:","[FileView] Warning: Could not find any SystemKeyboard objects.");
      }
   }

   /*** Cache commonly used icons ***/

   if (AccessObject(Self->ViewID, 5000, &view) IS ERR_Okay) {
      ActionTags(MT_CacheViewIcon, view, "folders/folder");
      ActionTags(MT_CacheViewIcon, view, "filetypes/empty");
      ReleaseObject(view);
   }

   Self->Flags |= FVF_TOTALREFRESH;
   acRefresh(Self);

   SubscribeTimer(Self->Head.UniqueID, Self->RefreshRate * 1000); /* Check for file changes every X number of seconds */

   return(ERR_Okay);
}

/*****************************************************************************

###ACTION###
Name:  LostFocus
Short: Deactivates keyboard monitoring (by force).
###END###

*****************************************************************************/

static ERROR FILEVIEW_LostFocus(struct FileView *Self, APTR Void)
{
   OBJECTPTR input;

   DPrintF("~LostFocus()","[FileView:%d]", Self->Head.UniqueID);

   if (Self->KeyboardChannel IS TRUE) {
      if (AccessObject(Self->InputID, 5000, &input) IS ERR_Okay) {
         UnsubscribeChannel(input, Self->Head.UniqueID);
         Self->KeyboardChannel = FALSE;
         ReleaseObject(input);
      }
      else ObjectError(Self, ERH_ActionNotify, ERR_AccessObject);
   }

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  MoveFilesTo
Short: Moves selected files to a destination location.

###DESCRIPTION###
<p>This method moves all selected files in the fileview to a destination
directory of your choosing.  If no files are selected in the fileview, this
method does nothing.  The MoveFilesTo method may return immediately following
the initial call and then move the files in the background.  If a failure
occurs, the user will be notified with an error dialog.</p>

###INPUT###
Dest: The directory that you want to copy the selected files to.

###SEE ALSO###
Method: MoveFilesTo

###END###

*****************************************************************************/

static ERROR FILEVIEW_MoveFilesTo(struct FileView *Self, struct mtMoveFilesTo *Args)
{
   ERROR error;

   if ((!Args) OR (!Args->Dest) OR (!Args->Dest[0])) return(ERR_Args);

   if ((error = FILEVIEW_CutFiles(Self, NULL)) IS ERR_Okay) {
      error = PasteTo(Self, Args->Dest);
      acRefresh(Self);
   }

   ActionMsg(MT_SelectNone, Self->ViewID, NULL);

   return(error);
}

/*****************************************************************************
** FileView: NewObject()
*/

static ERROR FILEVIEW_NewObject(struct FileView *Self, APTR Void)
{
   Self->RefreshRate = 0.25;
   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  ParentDirectory()
Short: Jumps to the parent directory of a FileView's current location.

###DESCRIPTION###
When this method is called, the FileView will jump to the parent directory
of the current location.  For instance, if the user is currently analysing
the 'athene:system/classes/' directory, then as a result of calling this method
the view will change to the 'athene:system/' directory.</p>

<p>If the current location is already at the root level ':', then this method
will return immediately as there is nothing above root.</p>

###RESULT###
ERR_Okay:   The method executed successfully, or the location is already set to root level.
ERR_Memory: The method ran out of memory during normal processing.

###END###

*****************************************************************************/

static ERROR FILEVIEW_ParentDirectory(struct FileView *Self, APTR Void)
{
   STRING NewLocation;
   LONG i, mark;

   if ((!Self->Location[0]) OR (Self->Location[0] IS ':')) return(ERR_Okay);

   mark = 0;
   for (i=0; (Self->Location[i] != 0); i++) {
      if ((Self->Location[i] IS '/') OR (Self->Location[i] IS ':')) mark = i;
   }

   if (mark) i = mark;

   if (Self->Location[i] IS 0) {
      return(SetFields(Self, FID_Location|TSTRING, NULL, TAGEND));
   }
   else {
      i--;
      while (i > 0) {
         if ((Self->Location[i] IS '/') OR (Self->Location[i] IS ':')) {
            if ((NewLocation = StrClone(Self->Location))) {
               NewLocation[i+1] = 0;
               SetFields(Self, FID_Location|TSTRING, NewLocation, TAGEND);
               FreeMemory(NewLocation);
               return(ERR_Okay);
            }
            else return(ERR_Memory);
         }
         i--;
      }

      return(SetFields(Self, FID_Location|TSTRING, NULL, TAGEND));
   }
}

/*****************************************************************************

###METHOD###
Name:  PasteFiles
Short: Pastes files from the clipboard to the current path of the file view.

###DESCRIPTION###
<p>Call the PasteFiles method to copy files from the clipboard into the
current path of the file view.  The file view will automatically refresh
itself so that pasted files are immediately displayed.</p>

###END###

*****************************************************************************/

static ERROR FILEVIEW_PasteFiles(struct FileView *Self, APTR Void)
{
   ERROR error;

   DPrintF("~PasteFiles()","[FileView:%d]", Self->Head.UniqueID);
   error = PasteTo(Self, Self->Location);
   StepBack();
   return(error);
}

/*****************************************************************************

###ACTION###
Name:  Refresh
Short: Refreshes a FileView's directory list.

###DESCRIPTION###
<p>When the Refresh action is called on a FileView object, the directory
that the FileView is monitoring will be read from scratch, and the list of
directories and files will be subsequently refreshed.</p>

###RESULT###
ERR_Okay:         The refresh was successful.
ERR_FieldNotSet:  The View field is not set.
ERR_CreateObject: An error occurred while creating a File object.

###END###

*****************************************************************************/

static ERROR FILEVIEW_Refresh(struct FileView *Self, APTR Void)
{
   struct DirInfo *info;
   STRING str, colstr;
   OBJECTPTR view, xml;
   UBYTE buffer[2048];
   LONG flags;
   LARGE systime, currenttime;
   BYTE dirchange;

   if (Self->Flags & FVF_TOTALREFRESH) {
      Self->Flags &= ~FVF_TOTALREFRESH;
   }
   else {
      /* Increase the rate of refresh to quickly scan the whole directory.
      ** The Timer action will reset the timer on completion.
      */

      DPrintF("2Refresh()","[FileView:%d] Location: %s", GetUniqueID(Self), Self->Location);

      SubscribeTimer(Self->Head.UniqueID, 20);
      Self->ResetTimer = TRUE;

      /*** Reset the directory scanner so that it starts from the beginning ***/

      CloseDirectory(Self->Dir);
      Self->Dir = NULL;

      return(ERR_Okay);
   }

   if (Self->Refresh IS TRUE) return(ERR_Okay|ERF_Notified); /* Do not allow a 'refresh within a refresh' */

   DPrintF("~2Refresh()","[FileView:%d] Location: %s", GetUniqueID(Self), Self->Location);

   if (!Self->ViewID) {
      DPrintF("@Refresh:","[FileView] The View field has not been set.");
      StepBack();
      return(ERR_FieldNotSet);
   }

   /*** Gain access to the View object for maximum communication speed ***/

   if (AccessObject(Self->ViewID, 5000, &view) != ERR_Okay) {
      return(StepObjectError(Self, ERH_Refresh, ERR_AccessObject));
   }

   if (GetField(view, FID_XML, FT_POINTER, &xml) != ERR_Okay) {
      ReleaseObject(view);
      return(StepObjectError(Self, ERH_Refresh, ERR_GetField));
   }

   if (Self->Dir) {
      CloseDirectory(Self->Dir);
      Self->Dir = NULL;
   }

   /*** Get a new file list ***/

   flags = RDF_READALL|RDF_QUALIFY|RDF_TAGS;
   if (Self->Flags & FVF_NOFILES) flags &= ~RDF_FILE; /* Do not read files */

   Self->Refresh = TRUE;

   acClear(view);

   /*** Define new column settings (if necessary) ***/

   if ((Self->Location[0] IS ':') OR (!Self->Location[0])) {
      str = "default(text:Name, len:160, showicons); freespace(text:Free Space, len:70, type:bytesize, rightalign); totalsize(text:Total Size, len:70, type:bytesize, rightalign)";
   }
#ifdef __linux__
   else str = "default(text:Name, len:160, showicons); size(text:Size, len:70, type:bytesize, rightalign); date(text:Date, len:100, type:date); permissions(text:Permissions,len:70); owner(text:Owner,len:50); group(text:Group,len:50)";
#else
   else str = "default(text:Name, len:160, showicons); size(text:Size, len:70, type:bytesize, rightalign); date(text:Date, len:100, type:date)";
#endif

   if ((GetField(view, FID_Columns, FT_STRING, &colstr) IS ERR_Okay) AND (colstr)) {
      if (StrCompare(colstr, str, 0, STR_MATCHLENGTH) != ERR_Okay) SetField(view, FID_Columns, FT_STRING, str);
   }
   else SetField(view, FID_Columns, FT_STRING, str);

   if ((Self->Flags & FVF_SHOWPARENT) AND (Self->Location[0] != ':') AND (Self->Location[0])) {
      UBYTE buffer[400];
      StrCopy("<directory icon=\"folders/folder\" sort=\"\001...\" insensitive>...</directory>", buffer, sizeof(buffer));
      acDataChannel(xml, Self->Head.UniqueID, DATA_XML, 1, buffer, StrLength(buffer) + 1, 1);
   }

   /* If the device is removeable, change the refresh timer so that we
   ** don't test the device so often.   This is useful because some badly
   ** written device drivers and hardware may presistently test themselves
   ** when there is no media in the drive.
   */

   struct DeviceInfo device;
   if (GetDeviceInfo(Self->Location, &device, sizeof(device)) IS ERR_Okay) {
      if ((device.Flags & DEVICE_REMOVEABLE)) {
         if (!(device.Flags & DEVICE_WRITE)) {
            /*** Read only device - do not refresh ***/
            UnsubscribeTimer(Self->Head.UniqueID);
         }
         else SubscribeTimer(Self->Head.UniqueID, 2000); /* Write/Read access to removeable device */
      }
      else SubscribeTimer(Self->Head.UniqueID, Self->RefreshRate * 1000);
   }

   systime = SystemTime();
   dirchange = FALSE;
   if (OpenDirectory(Self->Location, flags, &info) IS ERR_Okay) {
      while (ScanDirectory(info) IS ERR_Okay) {

         /* Some drives that have slow read access (like floppies) will take a long time
         ** to feed us information during this loop.  For this reason we call ProcessMessages()
         ** prevent the application from becoming frozen.  The location check is in case
         ** the user changes the focus of the FileView during the ProcessMessages() call.
         */

         StrCopy(Self->Location, buffer, sizeof(buffer));

         ProcessMessages(NULL, NULL);

         if (StrCompare(Self->Location, buffer, 0, STR_MATCHLENGTH) != ERR_Okay) {
            dirchange = TRUE;
            break;
         }

         AddFileItem(Self, xml, info);

         currenttime = SystemTime();
         if ((currenttime - systime) > 200) {
            systime = currenttime;
            acRefresh(view);
         }
      }

      CloseDirectory(info);
   }
   else DPrintF("Refresh:","[FileView] Failed to open directory \"%s\"", Self->Location);

   if (dirchange) {
      acClear(view);
   }
   else {
      acSort(view);
      acRefresh(view);
   }

   Self->Refresh = FALSE;

   ReleaseObject(view);

   if (dirchange) DelayMsg(AC_Refresh, Self->Head.UniqueID, NULL);

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************

###METHOD###
Name:  RenameFile
Short: Renames the most recently selected item in the file view.

###DESCRIPTION###
<p>This method simplifies the process of renaming the most recently selected
file for the user.  The method will pop-up an input box to ask the user for
the new file name.  If the user responds positively, the method will proceed
to rename the file.</p>

<p>If no file is selected, this method does nothing.</p>

###RESULT###
ERR_Okay: The process executed successfully.
ERR_AccessObject: Failed to access the referenced View object.

###END###

*****************************************************************************/

static ERROR FILEVIEW_RenameFile(struct FileView *Self, APTR Void)
{
   STRING selection;
   OBJECTPTR dialog;
   OBJECTID dialogid;
   ERROR error;

   if ((!Self->Location[0]) OR (Self->Location[0] IS ':')) return(ERR_Okay);

   if (GET_Selection(Self, &selection) != ERR_Okay) return(ERR_Okay);
   if ((!selection) OR (!selection[0])) return(ERR_Okay);

   DPrintF("~RenameFile()","[FileView:%d] %s", Self->Head.UniqueID, selection);

   if (CreateObject(ID_DIALOG, NF_CHILD, &dialog, &dialogid,
      FID_Image|TSTRING,     "icons:tools/edit(48)",
      FID_Buttons|TSTRING,   "okay;cancel",
      FID_Title|TSTRING,     TITLE_RENAME,
      FID_Flags|TLONG,       DF_INPUT|DF_INPUTREQUIRED,
      FID_String|TSTRING,    "Please enter the new name for the selected file or directory.",
      FID_TextInput|TSTRING, selection,
      FID_Action|TUNLISTED,  "Rename",
      TAGEND) IS ERR_Okay) {

      if (GET_SelectionPath(Self, &selection) IS ERR_Okay) {
         SetUnlistedField(dialog, "Src", selection);
         SubscribeField(dialog, FID_Response, Self);
         acShow(dialog);
         error = ERR_Okay;
      }
      else error = ERR_Failed;

      ReleaseObject(dialog);
   }
   else error = ERR_CreateObject;

   if (error != ERR_Okay) {
      if (dialogid) acFreeID(dialogid);
      ObjectError(Self, MT_RenameFile, error);
   }

   StepBack();
   return(error);
}

/*****************************************************************************
** Action: Timer
*/

static ERROR FILEVIEW_Timer(struct FileView *Self, struct acTimer *Args)
{
   OBJECTPTR xml, view;
   struct DirInfo *info;
   struct XMLTag **taglist, *tag, *scan;
   STRING compare;
   LARGE timestamp, size;
   BYTE filecount, newfile, refresh, sort;
   UBYTE buffer[512];
   LONG i, j;

   if (!Args) return(ERR_Args);

   /* Do nothing if the fileview is currently in the process of
   ** refreshing itself.
   */

   if (Self->Refresh) {
      return(ERR_Okay);
   }

   if (!Self->Dir) {
      if (OpenDirectory(Self->Location, RDF_FILES|RDF_DIRECTORIES|RDF_QUALIFY|RDF_PERMISSIONS|RDF_DATE|RDF_SIZE|RDF_TAGS, &Self->Dir) != ERR_Okay) {
         return(ERR_Okay);
      }
   }

   refresh = FALSE;
   sort = FALSE;
   if (AccessObject(Self->ViewID, 5000, &view) IS ERR_Okay) {
      if (GetFields(view, FID_XML|TPTR,  &xml,
                          FID_Tags|TPTR, &taglist,
                          TAGEND) != ERR_Okay) {
         taglist = NULL;
      }

      info = Self->Dir;
      for (filecount=0; filecount < 5; filecount++) {
         if (ScanDirectory(info) != ERR_Okay) {
            /*** We have reached the end of the directory - reset the scan ***/
            CloseDirectory(Self->Dir);
            Self->Dir = NULL;
            if (Self->ResetTimer) SubscribeTimer(Self->Head.UniqueID, Self->RefreshRate * 1000);
            break;
         }

         /*** Ignore hidden assigns ***/

         if ((info->Flags & RDF_ASSIGN) AND (info->Flags & RDF_HIDDEN)) continue;

         if (!taglist) {
            /* The view is empty and this is the first file
            ** to appear in the directory.
            */

            AddFileItem(Self, xml, info);
            acRefresh(view);

            if (GetField(xml, FID_Tags, FT_POINTER, &taglist) != ERR_Okay) taglist = NULL;

            break;
         }
         else {
            /* Find the file in the XML.  If we don't find it then
            ** the file is new and needs to be added.
            */

            newfile = TRUE;
            for (tag=taglist[0]; tag; tag=tag->Next) {
               if ((compare = ExtractFilename(tag))) {

                  if (StrCompare(info->Name, compare, 0, STR_MATCHLENGTH|STR_MATCHCASE) IS ERR_Okay) {
                     /* File found - extract the timestamp and test it against the file.
                     ** Note that we don't really need to test any other attributes, because
                     ** if the user changes the file size or permission details, programs
                     ** should update the timestamp to reflect those differences (and if it doesn't,
                     ** to hell with it :-)
                     */

                     timestamp = 0;
                     size = 0;
                     for (scan=tag->Child; scan; scan=scan->Next) {
                        if (StrCompare("date", scan->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                           for (i=1; i < scan->TotalAttributes; i++) {
                              if (StrCompare("sort", scan->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                                 timestamp = StrToInt(scan->Attributes[i].Value);
                                 break;
                              }
                           }
                        }
                        else if (StrCompare("size", scan->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                           for (i=1; i < scan->TotalAttributes; i++) {
                              if (StrCompare("sort", scan->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                                 size = StrToInt(scan->Attributes[i].Value);
                                 break;
                              }
                           }
                        }
                     }

                     if (((timestamp) AND (info->TimeStamp != timestamp)) OR (size != info->Size)) {
                        DPrintF("FileView:","File attribute changed \"%s\"", info->Name);

                        /*** Update the size and date tags ***/

                        for (scan=tag->Child; scan; scan=scan->Next) {
                           if (StrCompare("date", scan->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                              for (i=1; i < scan->TotalAttributes; i++) {
                                 if (StrCompare("sort", scan->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                                    if (info->Flags & RDF_DIRECTORY) StrFormat(buffer, sizeof(buffer), "D%014.0f", info->TimeStamp);
                                    else StrFormat(buffer, sizeof(buffer), "F%014.0f", info->TimeStamp);
                                    j = scan->Index;
                                    ActionTags(MT_SetXMLAttribute, xml, scan->Index, i, NULL, buffer);
                                    //ActionTags(MT_SetViewItem, view, -1, scan->Index, "date", "sort", buffer);

                                    scan = taglist[j]; /* Regain the address because SetXMLAttribute() invalidates it */
                                    if (scan->Child) {
                                       StrFormat(buffer, sizeof(buffer), "%04d%02d%02d %02d:%02d:%02d", info->Time.Year, info->Time.Month, info->Time.Day, info->Time.Hour, info->Time.Minute, info->Time.Second);
                                       ActionTags(MT_SetXMLAttribute, xml, scan->Child->Index, 0, NULL, buffer);
                                       //ActionTags(MT_SetViewItem, view, -1, scan->Child->Index, NULL, NULL, buffer);
                                       scan = taglist[j];
                                    }
                                 }
                              }
                           }
                           else if (StrCompare("size", scan->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                              for (i=1; i < scan->TotalAttributes; i++) {
                                 if (StrCompare("sort", scan->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                                    StrFormat(buffer, sizeof(buffer), "%014.0f", (DOUBLE)info->Size);
                                    j = scan->Index;
                                    ActionTags(MT_SetXMLAttribute, xml, scan->Index, i, NULL, buffer);
                                    //ActionTags(MT_SetViewItem, view, -1, scan->Index, "size", "sort", buffer);

                                    scan = taglist[j]; /* Regain the address because SetXMLAttribute() invalidates it */
                                    if (scan->Child) {
                                       StrFormat(buffer, sizeof(buffer), "%.0f", (DOUBLE)info->Size);
                                       ActionTags(MT_SetXMLAttribute, xml, scan->Child->Index, 0, NULL, buffer);
                                       //ActionTags(MT_SetViewItem, view, -1, scan->Child->Index, NULL, NULL, buffer);
                                    }
                                 }
                              }
                           }
                        }

                        refresh = TRUE;
                     }

                     newfile = FALSE;
                     break;
                  }
               }
            }

            if (newfile) {
               /*** The file does not exist in XML, so add it and re-sort the view ***/
               AddFileItem(Self, xml, info);
               if (GetField(xml, FID_Tags, FT_POINTER, &taglist) != ERR_Okay) taglist = NULL;
               sort = TRUE;
               refresh = TRUE;
            }
         }
      }

      /*** Now check for deleted files ***/

      if (taglist) {
         BYTE delete_count, state;
         ERROR error;

         if (Self->Location[0] IS ':') {
            j = 0;
            buffer[0] = 0;
         }
         else j = StrCopy(Self->Location, buffer, sizeof(buffer));

         delete_count = 0;
restart:
         /*** Figure out our tag position since our last delete ***/

         tag = taglist[0];
         for (i=0; (i < Self->DeleteIndex) AND (tag); i++) tag = tag->Next;

         if (!tag) {
            Self->DeleteIndex = 0;
            tag = taglist[0];
         }

         for (; (delete_count < 10) AND (tag); Self->DeleteIndex++, tag = tag->Next) {
            delete_count++;

            if (tag->Child) {
               if (StrCompare("...", tag->Child->Attributes[0].Value, 0, STR_MATCHLENGTH|STR_MATCHCASE) IS ERR_Okay) continue;

               /*** Extract the full name of the file for analysis ***/

               StrCopy(ExtractFilename(tag), buffer + j, sizeof(buffer) - j);

               state = FileCaseSensitive(TRUE);
               error = AnalyseLocation(buffer, NULL);
               FileCaseSensitive(state);

               if (error != ERR_Okay) {
                  DPrintF("FileView:","Detected deleted file \"%s\"", buffer);
                  ActionTags(MT_RemoveViewTag, view, tag->Index);
                  refresh = TRUE;
                  if (GetField(xml, FID_Tags, FT_POINTER, &taglist) != ERR_Okay) break;
                  goto restart; /* Restart so that we can re-establish our delete point */
               }
            }
         }
      }

      if (sort)    acSort(view);
      if (refresh) acRefresh(view);

      ReleaseObject(view);
      return(ERR_Okay);
   }
   else return(ObjectError(Self, ERH_Timer, ERR_AccessObject));
}

/*****************************************************************************

###METHOD###
Name:  ViewFiles
Short: Runs the view command for all currently selected files.

###DESCRIPTION###
<p>This method simplifies the process of viewing selected files for the user.
The routine utilises the Run class with the VIEW mode option to launch the files
in their respected viewers.</p>

<p>If no files are selected, this method does nothing.</p>

###RESULT###
ERR_Okay: The process executed successfully.
ERR_AccessObject: Failed to access the referenced View object.

###END###

*****************************************************************************/

static ERROR FILEVIEW_ViewFiles(struct FileView *Self, APTR Void)
{
   OBJECTPTR run, xml, view;
   struct XMLTag **taglist;
   LONG *tags;
   UBYTE buffer[300];
   LONG i;

   DPrintF("~ViewFiles()","[FileView:%d]", Self->Head.UniqueID);

   if ((!Self->Location[0]) OR (Self->Location[0] IS ':')) return(ERR_Okay);

   if (AccessObject(Self->ViewID, 5000, &view) IS ERR_Okay) {
      if ((GetFields(view, FID_SelectedTags|TPTR, &tags,
                           FID_XML|TPTR,          &xml,
                           FID_Tags|TPTR,         &taglist,
                           TAGEND) IS ERR_Okay) AND (tags)) {

         for (i=0; tags[i] != -1; i++) {
            StrFormat(buffer, sizeof(buffer), "%s%s", Self->Location, ExtractFilename(taglist[tags[i]]));

            CreateObject(ID_RUN, NULL, &run, NULL,
               FID_Mode|TSTRING,     "view",
               FID_Location|TSTRING, buffer,
               TAGEND);

            acActivate(run);
            acFree(run);
         }
      }
      ReleaseObject(view);
   }
   else return(ObjectError(Self, MT_ViewFiles, ERR_AccessObject));

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Filter
Short:  Defines a file filter through the use of wildcards.
Type:   STRING
Status: Read/Write

###DESCRIPTION###
<p>To define a file filter, set this field using standard wild-card values.
Both the asterisk and question-mark characters are accepted as wild-wards,
while the OR operator is reserved for use in a future update.  Here are some
filter examples:</p>

<pre>
   DML files:                             *.dml
   Files with extensions:                 *.*
   DML files with short names:            ???.dml
   Files where 'b' is a second character: ?b*
   Files starting with a, ending in b:    a*b
</pre>

<p>File filters are not case sensitive.</p>

###END###

*****************************************************************************/

static ERROR GET_Filter(struct FileView *Self, STRING *Value)
{
   if ((*Value = Self->Filter)) return(ERR_Okay);
   else return(ERR_NoData);
}

static ERROR SET_Filter(struct FileView *Self, STRING Value)
{
   LONG i;

   DPrintF("~Set.Filter:","[FileView:%d] %s", Self->Head.UniqueID, Value);

   if (!Value) Self->Filter[0] = 0;
   else {
      for (i=0; (i < sizeof(Self->Filter)-1) AND (Value[i]); i++) Self->Filter[i] = Value[i];
      Self->Filter[i] = 0;
   }

   if (Self->Head.ObjectFlags & NF_INITIALISED) {
      acRefresh(Self);
   }

   StepBack();
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Location
Short:  Reflects the directory location that a fileview is analysing.
Type:   STRING
Status: Get/Set

###DESCRIPTION###
<p>If you want a fileview to analyse a specific directory, writing to
this field will force the object to switch to the new location and refresh
the file list.</p>

<p>To change to the root directory, set the Location to a zero length string,
or write the field with a NULL pointer.</p>

###END###

*****************************************************************************/

static ERROR GET_Location(struct FileView *Self, STRING *Value)
{
   if ((*Value = Self->Location)) return(ERR_Okay);
   else return(ERR_NoData);
}

static ERROR SET_Location(struct FileView *Self, STRING Value)
{
   WORD i, len;

   /*** Check if the new string is any different to the current location ***/

   if ((!Self->Location[0]) OR (Self->Location[0] IS ':')) {
      if ((!Value) OR (!Value[0]) OR (Value[0] IS ':')) return(ERR_Okay);
   }

   if (StrCompare(Value, Self->Location, 0, STR_MATCHLENGTH) IS ERR_Okay) return(ERR_Okay);

   /*** Set the new location string ***/

   if ((!Value) OR (!Value[0])) {
      Self->Location[0] = 0;
   }
   else {
      for (len=0; Value[len]; len++);
      while ((len > 0) AND (Value[len-1] != '/') AND (Value[len-1] != '\\') AND (Value[len-1] != ':')) len--;

      for (i=0; (i < len) AND (i < sizeof(Self->Location)-1); i++) Self->Location[i] = Value[i];
      Self->Location[i] = 0;
   }

   /* If the object is initialised, update the location
   ** string and switch to the new directory.
   */

   if (Self->Dir) {
      CloseDirectory(Self->Dir);
      Self->Dir = NULL;
   }

   if (Self->Head.ObjectFlags & NF_INITIALISED) {
      Self->Flags |= FVF_TOTALREFRESH;
      if (acRefresh(Self) != ERR_Okay) {
         return(ERR_Refresh);
      }
   }

   DPrintF("Set.Location:","[FileView] = %s", Self->Location);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Selection
Short:  Readable field that indicates the name of the currently selected item.
Type:   STRING
Status: Get

###DESCRIPTION###
<p>The Selection field provides a method for learning what the currently
selected file or directory is.  If no selection is active then this
function will return a value of NULL.  The path leading to the file <i>will
not</i> be included in the resulting string.  If the current selection is a
directory or assignment, any trailing symbols will be stripped from the end
of the directory name.</p>

<p>The Selection is returned exactly as it appears to the user - so if you
have opted to strip all extensions from file names, you will not get the
'exact' filename as it is described on the filesystem.</p>

###SEE ALSO###
Fields: SelectionFile, SelectionPath

###END###

*****************************************************************************/

static ERROR GET_Selection(struct FileView *Self, STRING *Value)
{
   struct XMLTag **tags;
   OBJECTPTR view, xml;
   LONG tagindex;
   WORD i;

   *Value = NULL;

   if (AccessObject(Self->ViewID, 5000, &view) IS ERR_Okay) {
      if ((GetFields(view, FID_XML|TPTR,          &xml,
                           FID_SelectedTag|TLONG, &tagindex,
                           FID_Tags|TPTR,         &tags,
                           TAGEND) IS ERR_Okay) AND (tagindex != -1) AND (xml)) {

         if (tags[tagindex]->Child) StrCopy(tags[tagindex]->Child->Attributes[0].Value, Self->Selection, sizeof(Self->Selection));
         else StrCopy(ExtractFilename(tags[tagindex]), Self->Selection, sizeof(Self->Selection));

         /*** Strip any trailing directory symbols ***/

         for (i=0; Self->Selection[i]; i++);
         while ((i > 0) AND ((Self->Selection[i-1] IS ':') OR (Self->Selection[i-1] IS '/'))) i--;
         Self->Selection[i] = 0;

         *Value = Self->Selection;
      }
      ReleaseObject(view);
   }
   else return(ERR_AccessObject);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   SelectionFile
Short:  The currently selected file or directory (fully-qualified).
Type:   STRING
Status: Get

###DESCRIPTION###
<p>The SelectionFile field provides a method for learning what the currently
selected file or directory is.  If no selection is active then this function
will return a value of NULL.  The path leading to the file <i>will not</i>
be included in the resulting string.  If the current selection is a
directory or assignment, the correct trailing symbol will be appended to
the end of the string to indicate the file type.</p>

###END###

*****************************************************************************/

static ERROR GET_SelectionFile(struct FileView *Self, STRING *Value)
{
   struct XMLTag **tags;
   OBJECTPTR view, xml;
   LONG tagindex;

   *Value = NULL;

   if (AccessObject(Self->ViewID, 5000, &view) IS ERR_Okay) {
      if ((GetFields(view, FID_XML|TPTR,          &xml,
                           FID_SelectedTag|TLONG, &tagindex,
                           FID_Tags|TPTR,         &tags,
                           TAGEND) IS ERR_Okay) AND (tagindex != -1) AND (xml)) {

         StrCopy(ExtractFilename(tags[tagindex]), Self->Selection, sizeof(Self->Selection));

         *Value = Self->Selection;
      }
      ReleaseObject(view);
   }
   else return(ERR_AccessObject);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   SelectionPath
Short:  The complete file-path of the current selection.
Type:   STRING
Status: Get

###DESCRIPTION###
<p>The SelectionPath field provides a method for learning what the currently
selected file or directory is.  If no selection is active then this function
will return a value of NULL.  The path leading to the file will be included
in the resulting string.  The path is fully-qualified, so a trailing slash
or colon will be present in the event that the selection is a directory or
assignment.</p>

<p>The resulting string will remain valid until the fileview object is freed
or the SelectionPath is used a consecutive time.</p>

###END###

*****************************************************************************/

static ERROR GET_SelectionPath(struct FileView *Self, STRING *Value)
{
   struct XMLTag **tags;
   OBJECTPTR view, xml, context;
   UBYTE buffer[500];
   LONG tagindex;
   ERROR error;

   *Value = NULL;

   if (AccessObject(Self->ViewID, 5000, &view) IS ERR_Okay) {
      if ((GetFields(view, FID_XML|TPTR,          &xml,
                           FID_SelectedTag|TLONG, &tagindex,
                           FID_Tags|TPTR,         &tags,
                           TAGEND) IS ERR_Okay) AND (tagindex != -1) AND (xml)) {

         StrFormat(buffer, sizeof(buffer), "%s%s", Self->Location, ExtractFilename(tags[tagindex]));

         ReleaseObject(view);

         if (Self->SelectionPath) {
            FreeMemory(Self->SelectionPath);
            Self->SelectionPath = NULL;
         }

         context = SetContext(Self);
         error = AllocMemory(StrLength(buffer)+1, MEM_STRING|MEM_NOCLEAR, &Self->SelectionPath, NULL);
         SetContext(context);

         if (error IS ERR_Okay) {
            StrCopy(buffer, Self->SelectionPath, COPY_ALL);
            *Value = Self->SelectionPath;
            return(ERR_Okay);
         }
         else return(ObjectError(Self, ERH_GetField, ERR_AllocMemory));
      }
      else {
         ReleaseObject(view);
         return(ERR_FieldNotSet);
      }
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));
}

/*****************************************************************************

###FIELD###
Name:   View
Short:  Must refer to a GUI object that will provide viewing services.
Type:   OBJECTID
Status: Read/Init

###DESCRIPTION###
<p>Because a FileView object does not draw its own graphics, you need to
provide it with a reference to a GUI object that is capable of providing
an interactive viewing service.  To do so, you need to create the viewing
object and then refer to it by writing its ID to this field.</p>

<p>Currently we recommend that you use a Text object to provide viewing
services, but any GUI class can be used so long as it can draw graphics and
provide a functional user interaction service.</p>

###END###

*****************************************************************************/

/*****************************************************************************
** Internal: InitKeyboard()
*/

static void InitKeyboard(struct FileView *Self)
{
   OBJECTPTR input;

   if (Self->KeyboardChannel IS FALSE) {
      if (AccessObject(Self->InputID, 5000, &input) IS ERR_Okay) {
         SubscribeChannel(input, Self);
         Self->KeyboardChannel = TRUE;
         ReleaseObject(input);
      }
   }
}

/******************************************************************************
** Internal: AddFileItem()
*/

static void AddFileItem(struct FileView *Self, OBJECTPTR XML, struct DirInfo *Info)
{
   struct DeviceInfo device;
   struct FileTime *time;
   STRING user, group;
   UBYTE buffer[1024], total_size[50], free_space[50], filename[256], strpermissions[24];
   WORD display, i;
   STRING icon;

   //DebugState(FALSE, -1);

   buffer[0] = 0;
   if (Info->Flags & RDF_ASSIGN) {
      STRING label;

      /*** Get the icon to use for displaying the assign ***/

      if (!(Info->Flags & RDF_HIDDEN)) {
         if (!(icon = GetFileIcon(Info->Name))) icon = "folders/folder";

         /*** Determine the size and available space on the device ***/

         total_size[0] = 0;
         free_space[0] = 0;
         if (GetDeviceInfo(Info->Name, &device, sizeof(device)) IS ERR_Okay) {
            if (device.Flags & (DEVICE_FLOPPYDISK|DEVICE_HARDDISK|DEVICE_COMPACTDISC)) {
               if (device.DeviceSize >= 1) StrFormat(total_size, sizeof(total_size), "%014.0f", device.DeviceSize);
               if (device.BytesAvailable >= 1) StrFormat(free_space, sizeof(free_space), "%014.0f", device.BytesAvailable);
            }
         }

         label = NULL;
         if (Info->Tags) {
            /*** Check for a label ***/
            for (i=0; Info->Tags[i]; i++) {
               if (StrCompare("LABEL:", Info->Tags[i], 6, NULL) IS ERR_Okay) {
                  label = Info->Tags[i] + 6;
               }
            }
         }

         if (label) {
            StrFormat(buffer, sizeof(buffer), "<directory icon=\"%s\" sort=\"\001%s\" insensitive name=\"%s\">%s (%s)<totalsize>%s</totalsize><freespace>%s</freespace></directory>", icon, Info->Name, Info->Name, Info->Name, label, total_size, free_space);
         }
         else StrFormat(buffer, sizeof(buffer), "<directory icon=\"%s\" sort=\"\001%s\" insensitive>%s<totalsize>%s</totalsize><freespace>%s</freespace></directory>", icon, Info->Name, Info->Name, total_size, free_space);
      }
   }
   else if (Info->Flags & RDF_DIRECTORY) {
      time = &Info->Time;
      if (time->Year) {
         i = StrFormat(buffer, sizeof(buffer), "<directory icon=\"folders/folder\" sort=\"\001%s\" insensitive>%s<date sort=\"D%012.0f\">%04d%02d%02d %02d:%02d:%02d</date>",
            Info->Name, Info->Name, (DOUBLE)Info->TimeStamp, time->Year, time->Month, time->Day, time->Hour, time->Minute, time->Second);
      }
      else i = StrFormat(buffer, sizeof(buffer), "<directory icon=\"folders/folder\" sort=\"\001%s\" insensitive>%s", Info->Name, Info->Name);

      convert_permissions(Info->Permissions, strpermissions);
      if (!(user = ConvertUserID(Info->UserID))) user = "";
      if (!(group = ConvertGroupID(Info->GroupID))) group = "";
      StrFormat(buffer+i, sizeof(buffer)-i, "<owner>%s</owner><group>%s</group><permissions>%s</permissions></directory>", user, group, strpermissions);
   }
   else if (Info->Flags & RDF_FILE) {
      if (Self->Filter[0]) {
         if (StrCompare(Self->Filter, Info->Name, 0, STR_WILDCARD) IS ERR_Okay) display = TRUE;
         else display = FALSE;
      }
      else display = TRUE;

      if (display IS TRUE) {
         StrFormat(buffer, sizeof(buffer), "%s%s", Self->Location, Info->Name);
         icon = GetFileIcon(buffer);

         if (Self->Flags & FVF_NOEXTENSIONS) {
            StrCopy(Info->Name, filename, sizeof(filename));
            StripExtension(filename);
            i = StrFormat(buffer, sizeof(buffer), "<file icon=\"%s\" name=\"%s\">%s<size sort=\"%014.0f\">%.0f</size><date sort=\"F%012.0f\">%04d%02d%02d %02d:%02d:%02d</date>", icon, Info->Name, filename, (DOUBLE)Info->Size, (DOUBLE)Info->Size, (DOUBLE)Info->TimeStamp, Info->Time.Year, Info->Time.Month, Info->Time.Day, Info->Time.Hour, Info->Time.Minute, Info->Time.Second);
         }
         else i = StrFormat(buffer, sizeof(buffer), "<file icon=\"%s\">%s<size sort=\"%014.0f\">%.0f</size><date sort=\"F%012.0f\">%04d%02d%02d %02d:%02d:%02d</date>", icon, Info->Name, (DOUBLE)Info->Size, (DOUBLE)Info->Size, (DOUBLE)Info->TimeStamp, Info->Time.Year, Info->Time.Month, Info->Time.Day, Info->Time.Hour, Info->Time.Minute, Info->Time.Second);

         convert_permissions(Info->Permissions, strpermissions);
         if (!(user = ConvertUserID(Info->UserID))) user = "";
         if (!(group = ConvertGroupID(Info->GroupID))) group = "";
         StrFormat(buffer+i, sizeof(buffer)-i, "<owner>%s</owner><group>%s</group><permissions>%s</permissions></file>", user, group, strpermissions);
      }
   }

   if (buffer[0]) acDataChannel(XML, Self->Head.UniqueID, DATA_XML, 1, buffer, StrLength(buffer) + 1, 1);

   //DebugState(TRUE, -1);
}

/******************************************************************************
** Internal: RenameFileItem()
*/

static ERROR RenameFileItem(struct FileView *Self, STRING Name, STRING NewName)
{
   OBJECTPTR xml, view;
   struct XMLTag **taglist, *tag;
   STRING compare;
   LONG i;

   DPrintF("RenameItem()","%s to %s", Name, NewName);

   if (AccessObject(Self->ViewID, 5000, &view) IS ERR_Okay) {
      if ((GetFields(view, FID_XML|TPTR,  &xml,
                           FID_Tags|TPTR, &taglist,
                           TAGEND) IS ERR_Okay) AND (taglist)) {

         for (tag=taglist[0]; tag; tag=tag->Next) {
            if ((compare = ExtractFilename(tag))) {
               if (StrCompare(Name, compare, 0, STR_MATCHLENGTH|STR_MATCHCASE) IS ERR_Okay) {
                  i = StrLength(NewName);
                  UBYTE buffer[i+2];

                  if (StrCompare("directory", tag->Attributes[0].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                     StrCopy(NewName, buffer, COPY_ALL);
                     buffer[i++] = '/';
                     buffer[i] = 0;
                     NewName = buffer;
                  }

                  /*** Update the real name ***/

                  for (i=1; i < tag->TotalAttributes; i++) {
                     if (StrCompare("name", tag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
                        ActionTags(MT_SetXMLAttribute, xml, tag->Index, 0, NULL, NewName);
                        break;
                     }
                  }

                  /*** Update the name by setting the content string ***/

                  ActionTags(MT_SetXMLAttribute, xml, tag->Child->Index, 0, NULL, NewName);

                  acSort(view);
                  acRefresh(view);
                  break;
               }
            }
         }
      }

      ReleaseObject(view);
      return(ERR_Okay);
   }
   else return(ERR_AccessObject);
}

static void StripExtension(STRING String)
{
   WORD i;
   for (i=0; String[i]; i++);
   while ((i > 0) AND (String[i] != '.')) i--;
   if (i > 0) String[i] = 0;
}

static STRING ExtractFilename(struct XMLTag *Tag)
{
   WORD i;

   if (!Tag) return(NULL);

   for (i=1; i < Tag->TotalAttributes; i++) {
      if (StrCompare("name", Tag->Attributes[i].Name, 0, STR_MATCHLENGTH) IS ERR_Okay) {
         return(Tag->Attributes[i].Value);
      }
   }

   if (Tag->Child) return(Tag->Child->Attributes[0].Value);

   return(NULL);
}

/*****************************************************************************
** Internal: PasteTo()
**
** Launches a separate process for pasting files to a destination.
*/

static ERROR PasteTo(struct FileView *Self, STRING Folder)
{
   OBJECTPTR run;
   UBYTE args[512];
   ERROR error;

   if (StrFormat(args, sizeof(args), "commands:pastefiles.dml -args dest=\"%s\"", Folder) < sizeof(args)-1) {
      if ((error = CreateObject(ID_RUN, NF_CHILD, &run, NULL,
         FID_Location|TSTRING, "bin:runscript",
         FID_Args|TSTRING,     args,
         TAGEND)) IS ERR_Okay) {
         error = acActivate(run);
         acFree(run);
      }

      if (error != ERR_Okay) {
         /*** If the pastefiles script failed, resort to direct clipboard access ***/
         error = ActionTags(MT_PasteClipboardFiles, glClipboard, Folder, NULL);
         acRefresh(Self);
      }
   }
   else error = ERR_BufferOverflow;

   return(error);
}

/*****************************************************************************
** Internal: convert_permissions()
** Short:    Translates permission flags into a readable string.
*/

static void convert_permissions(LONG Permissions, STRING Output)
{
   if (Permissions & PERMIT_READ)  *Output++ = 'r'; else *Output++ = '-';
   if (Permissions & PERMIT_WRITE) *Output++ = 'w'; else *Output++ = '-';
   if (Permissions & PERMIT_EXEC)  *Output++ = 'x'; else *Output++ = '-';

   *Output++ = ' ';

   if (Permissions & PERMIT_GROUP_READ)  *Output++ = 'r'; else *Output++ = '-';
   if (Permissions & PERMIT_GROUP_WRITE) *Output++ = 'w'; else *Output++ = '-';
   if (Permissions & PERMIT_GROUPID) *Output++ = 'g';
   else if (Permissions & PERMIT_GROUP_EXEC) *Output++ = 'x'; else *Output++ = '-';

   *Output++ = ' ';

   if (Permissions & PERMIT_OTHERS_READ)  *Output++ = 'r'; else *Output++ = '-';
   if (Permissions & PERMIT_OTHERS_WRITE) *Output++ = 'w'; else *Output++ = '-';
   if (Permissions & PERMIT_OTHERS_EXEC)  *Output++ = 'x'; else *Output++ = '-';

   if (Permissions & PERMIT_USERID) { *Output++ = ' '; *Output++ = 's'; }

   *Output++ = 0;
}
